package org.aohe.core.sane.utils;

import au.com.southsky.jfreesane.SaneDevice;
import au.com.southsky.jfreesane.SaneException;
import au.com.southsky.jfreesane.SaneOption;
import au.com.southsky.jfreesane.SaneSession;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import lombok.extern.slf4j.Slf4j;
import org.aohe.core.result.R;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.net.InetAddress;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import au.com.southsky.jfreesane.OptionValueType;
import org.aohe.core.sane.mode.ScanMode;

@Slf4j
public class SaneSessionUtils {
    
    private static final Object lock = new Object();
    
    private static volatile SaneSession saneSession;
    
    private static volatile SaneDevice saneDevice;
    
    private static final RemoteConfig remoteConfig = RemoteConfig.getInstance();

    /**
     * 获取一个SaneSession实例，用于与SANE设备进行通信
     * @return SaneSession实例
     * @throws IOException 如果获取SaneSession实例时发生I/O错误
     */
    public static SaneSession getSaneSession() throws IOException {
        log.info("saneSession状态：{}", saneSession != null);
        if (saneSession == null) {
            synchronized(lock){
                if (saneSession == null) {
                    log.info("jFreeSaneConfig:{}", JSON.toJSONString(remoteConfig));
                    InetAddress address = InetAddress.getByName(remoteConfig.getIp());
                    saneSession = SaneSession.withRemoteSane(address);
                    log.info("初始化saneSession完成，状态：{}", saneSession != null);
                }
            }
        }
        return saneSession;
    }
    
    /**
     * 关闭SaneSession对象，如果对象不为null
     */
    public static void closeSaneSession() {
        if (saneSession != null) {
            try {
                saneSession.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
            saneSession = null;
        }
    }
    
    /**
     * 获取所有的SANE驱动器设备列表
     *
     * @return SANE驱动器设备列表
     * @throws IOException    如果在通信中发生I/O错误
     * @throws SaneException  如果在与SANE交互中发生错误
     */
    public static List<SaneDevice> getSaneDrivers() throws IOException, SaneException {
        return getSaneSession().listDevices();
    }
    
    
    /**
     * 设置Sane设备
     * @param name 设备名称
     * @return 设备对象
     * @throws IOException 输入/输出异常
     * @throws SaneException Sane异常
     */
    public static SaneDevice setSaneDevice(String name) throws IOException, SaneException {
        if (saneDevice == null || !saneDevice.getName().equals(name)) {
            if (saneDevice != null && saneDevice.isOpen()) {
                // 关闭当前在使用的SANE设备
                saneDevice.close();
            }
            saneDevice = saneSession
                    .listDevices().stream().filter(s -> name.equals(s.getName())).findFirst().orElse(null);
        }
        return saneDevice;
    }
    
    /**
     * 获取Sane设备（如果不存在则返回null）
     * @return Sane设备
     */
    public static SaneDevice getDevice() {
        return saneDevice;
    }
    
    /**
     * 获取指定名称的Sane设备，如果设备不存在则返回null
     * @param name 设备名称
     * @return 指定名称的Sane设备
     * @throws RuntimeException 运行时异常
     */
    public static SaneDevice getDevice(String name) {
        try {
            if(StrUtil.isBlank(name)){
                return saneDevice;
            }
            return setSaneDevice(name);
        } catch (IOException | SaneException e) {
            throw new RuntimeException(e);
        }
    }
    
    
    /**
     * 获取一个正常的文件
     * 打开Sane设备
     * 通过Sane设备获取图像
     * 创建或更新最后访问时间的文件
     * 将图像写入文件
     * 返回文件
     * @throws RuntimeException - 如果发生I/O错误
     */
    public static File getSaneFile() throws IOException {
        try {
            saneDevice.open();
            BufferedImage bufferedImage =  saneDevice.acquireImage();
            File file = FileUtil.createTempFile("sane","png",true);
            ImageIO.write(bufferedImage, "png", file);
            return file;
        } catch (IOException | SaneException e) {
            throw new RuntimeException(e);
        }finally {
            if(saneDevice != null){
                saneDevice.close();
            }
        }
    }

    /**
     * 获取扫描仪参数
     * @return list
     * @throws IOException io异常
     * @throws SaneException 扫描异常
     */
    public static List<SaneOption> getOptions(String name){
        SaneDevice saneDevice = getDevice(name);
        if (saneDevice == null) {
            return null;
        }
        try {
            if (!saneDevice.isOpen()) {
                saneDevice.open();
            }
            return saneDevice.listOptions();
        } catch (IOException | SaneException e ) {
            log.error("系统错误 ", e);
            return null;
        }

    }

//    /**
//     * BW - 0
//     * COLOR - 2
//     * GRAYSCALE - 1
//     * true ->auto
//     * @param params
//     */
//    /**
//     * 设置参数
//     * @param params 参数 Dpi ->double 0-bw 2-color 1-gray true-auto
//     * @throws SaneException 系统错误
//     */
//    public static void setOption(JSONObject params) throws SaneException, IOException {
//        if(saneDevice == null){
//            throw new SaneException("please first set device");
//        }
//        if(params == null){
//            return;
//        }
//        Double key = params.getDouble("dpi");
//        if(key != null){
//            SaneOption dpi = saneDevice.getOption("Dpi");
//            dpi.setIntegerValue(key.intValue());
//        }
//
//        Integer mode = params.getInteger("mode");
//        if(mode != null){
//             saneDevice.getOption("model");
//            switch (mode){
//                case 0:
//                    source.setColor(Source.ColorMode.BW);
//                    break;
//                case 1:
//                    source.setColor(Source.ColorMode.GRAYSCALE);
//                    break;
//                case 2:
//                default:
//                    source.setColor(Source.ColorMode.COLOR);
//            }
//        }
//
//        Boolean paperMode = params.getBoolean("paperMode");
//        if(paperMode != null){
//            source.setAutoDocumentFeeder(paperMode);
//        }
//
//        Boolean systemUI = params.getBoolean("systemUI");
//        if(systemUI != null){
//            source.setSystemUI(systemUI);
//        }
//    }

    /**
     * 设置扫描仪
     * @param deviceName
     * @param options
     * @return
     * @throws IOException
     * @throws SaneException
     */
    private R doSetOptions(String deviceName, LinkedHashMap<String, Object> options) throws IOException, SaneException {
        SaneDevice saneDevice = getDevice(deviceName);

        if (saneDevice == null) {
            return null;
        }
        if (!saneDevice.isOpen()) {
            saneDevice.open();
        }
        for (Map.Entry<String, Object> entry : options.entrySet()) {
            String key = entry.getKey();
            Object value = entry.getValue();
            SaneOption option = saneDevice.getOption(key);
            if (ObjectUtil.isEmpty(option)) {
                String errorMessage = String.format("option %s is not null,请检查该属性值是否含有使用条件.", key);
                return R.fail(errorMessage);
            }
            OptionValueType optionType = option.getType();
            if (!option.isActive()) {
                String errorMessage2 = String.format("option %s is not active,请检查该属性值是否含有使用条件.%n option %s description:%s", option.getName(), option.getName(), option.getDescription());
                return R.fail(errorMessage2);
            }
            switch (optionType) {
                case BOOLEAN:
                    boolean val = Boolean.parseBoolean(String.valueOf(value));
                    option.setBooleanValue(val);
                    break;
                case INT:
                    try {
                        int intVal = Integer.parseInt(String.valueOf(value));
                        int max = option.getRangeConstraints().getMaximumInteger();
                        int min = option.getRangeConstraints().getMinimumInteger();
                        int quantum = option.getRangeConstraints().getQuantumInteger();
                        if (intVal % quantum != 0 || intVal > max || intVal < min) {
                            String errorMessage3 = String.format("无效的%s值%s,有效范围为:[%s-%s],Quantum:%s", key, value, Integer.valueOf(min), Integer.valueOf(max), Integer.valueOf(quantum));
                            log.error(errorMessage3);
                            return R.fail(errorMessage3);
                        }
                        option.setIntegerValue(intVal);
                        break;
                    } catch (Exception nfe) {
                        String errorMessage4 = String.format("错误的Int类型值%s", value);
                        log.error(errorMessage4, nfe);
                        return R.fail(errorMessage4);
                    }
                case FIXED:
                    try {
                        double doubleVal = Double.parseDouble(String.valueOf(value));
                        double max2 = BigDecimal.valueOf(option.getRangeConstraints().getMaximumFixed()).setScale(2, RoundingMode.HALF_UP).doubleValue();
                        double min2 = BigDecimal.valueOf(option.getRangeConstraints().getMinimumFixed()).setScale(2, RoundingMode.HALF_UP).doubleValue();
                        double quantum2 = BigDecimal.valueOf(option.getRangeConstraints().getQuantumFixed()).setScale(2, RoundingMode.HALF_UP).doubleValue();
                        if ((doubleVal / quantum2) % 1.0d != 0.0d || doubleVal > max2 || doubleVal < min2) {
                            String errorMessage5 = String.format("无效的%s值%s,有效范围为:[%s-%s],Quantum:%s", key, value, Double.valueOf(min2), Double.valueOf(max2), Double.valueOf(quantum2));
                            log.warn(errorMessage5);
                            return R.fail(errorMessage5);
                        }
                        option.setFixedValue(doubleVal);
                        break;
                    } catch (Exception nfe2) {
                        String errorMessage6 = String.format("错误的Double类型值%s", value);
                        log.error(errorMessage6, nfe2);
                        return R.fail(errorMessage6);
                    }
                case STRING:
                    try {
                        String stringVal = value.toString();
                        List<String> effectiveValues = option.getStringConstraints();
                        if (effectiveValues != null) {
                            if (!effectiveValues.contains(stringVal)) {
                                String errorMessage7 = String.format("无效的%s值%s,有效范围为:[%s]", key, value, String.join(",", effectiveValues));
                                log.warn(errorMessage7);
                                return R.fail(errorMessage7);
                            }
                            option.setStringValue(stringVal);
                            if (option.getName().equals(getModeString())) {
                                //SaneSessionUtil.scanMode = (ScanMode) Enum.valueOf(ScanMode.class, stringVal);
                            }
                        }
                        break;
                    } catch (Exception nfe3) {
                        String errorMessage8 = String.format("错误的String类型值%s", value);
                        log.error(errorMessage8, nfe3);
                        return R.fail(errorMessage8);
                    }
            }
        }
        return R.ok();
    }

    public static String getModeString() {
        return "mode";
    }



    /**
     * 释放资源
     */
    public static void resource(){
        try {
            if(saneDevice != null){
                closeSaneDevice();
            }
            if(saneSession != null){
                closeSaneSession();
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * 关闭Sane设备
     */
    public static void closeSaneDevice() throws IOException {
        if(saneDevice != null){
            try {
                saneDevice.close();
            } catch (IOException e) {
                throw new IOException(e);
            }
            saneDevice = null;
        }
        
    }



}

